package requestExecute;

import java.util.*;

public class Proxy
{
	private List<Coppia> toDo;
	private Map<Key, Result> results;

	public Proxy(int numExecutors)
	{
		toDo = new LinkedList<Coppia>();
		results = new HashMap<Key, Result>();
		for(int i=0;i<numExecutors;i++)
		{
			(new Thread(new Executor(toDo, results))).start();
		}
	}

	public Proxy()
	{
		this(5);
	}

	public Result syncRequest(Request r)
	{
		return syncRetrieve(asyncRequest(r), 0);
	}

	public Key asyncRequest(Request r)
	{
		Key k = new Key();
		Coppia c = new Coppia(k, r);

		synchronized(results)
		{
			results.put(k, new Result());
		}

		synchronized(toDo)
		{
			toDo.add(c);
			toDo.notify();
		}

		return k;
	}

	public Result syncRetrieve(Key k, long timeout)
	{
		Result r = null;
		synchronized(results)
		{
			r = results.get(k);
		}

		synchronized(r)
		{
			boolean repeat = true;
			while(!r.isValid() && repeat)
			{
				try
				{
					r.wait(timeout);
					repeat = false;
				}
				catch(InterruptedException ie)
				{
					ie.printStackTrace();
				}
			}
			if(!r.isValid())
			{
				r = null;
			}
		}
		return r;
	}

	public Result asyncRetrieve(Key k)
	{
		Result retval = null;
		synchronized(results)
		{
			retval = results.get(k);
		}

		/* non faccio il test per consistenz di comportamento con la versione sincrona
		if(retval == null)//non puo' succedere se ho gia' fatto la richiesta di servizio
		*/

		synchronized(retval)
		{
			if(!retval.isValid())
			{
				retval = null;
			}
		}
		return retval;
	}

	public boolean invalidate(Key k)
	{
		synchronized(results)
		{
			Result r = results.get(k);
			synchronized(r)
			{
				if(r.isValid())
				{
					results.remove(k);
					return true;
				}
			}
		}
		return false;
	}

	class Executor
	implements Runnable
	{
		List<Coppia> l;
		Map<Key, Result> m;
		private boolean go = true;

		Executor(List<Coppia> myL, Map<Key, Result> myM)
		{
			l = myL;
			m = myM;
		}

		void kill()
		{
			go = false;
		}

		public void run()
		{
			while(go)
			{
				Coppia c = null;
				synchronized(l)
				{
					while(l.isEmpty())
					{
						try
						{
							l.wait();
						}
						catch(InterruptedException ie)
						{
							ie.printStackTrace();
						}
					}
					c = l.remove(0);
				}

				Object res = c.r.execute();

				Result r = null;
				synchronized(results)
				{
					r = results.get(c.k);
				}
				synchronized(r)
				{
					r.r = res;
					r.valid = true;
					r.notifyAll();
				}
			}
		}
		
	}

	class Coppia
	{
		Key k;
		Request r;

		Coppia(Key myK, Request myR)
		{
			k = myK;
			r = myR;
		}
	}
}
